/*

  ECHOSERV.C
  ==========

  Simple TCP/IP echo server using globus I/O with
  authentication and encryption included.

  Usage: echoserver <port_number>
  
  Arto Ters <arto.teras@hip.fi> 18.6.2002

  Compare with a server using direct TCP socket calls, for example
  http://www.paulgriffiths.net/program/c/echoserv.html

*/

/* Set AUTHENTICATE to one to use GSS-API authentication. */
#define AUTHENTICATE 1

/* Set ENCRYPTION to one to use ssl */
#define ENCRYPTION 1

#include <stdlib.h>
#include <stdio.h>

#include <globus_common.h>
#include <globus_io.h>

/*  Global constants  */
#define MAX_LINE           (1000)

int main(int argc, char *argv[]) {
    unsigned short int port;         /* Port number to listen           */
    char      buffer[MAX_LINE];      /* Buffer for the string to echo   */
    char      *endptr;               /* For strtol()                    */
    int       error;                 /* Helper variable for errors      */

    /* Variables for Globus functions */
    int                rval;         /* Return value (old APIs)         */
    globus_result_t    rptr;         /* Return value (new APIs)         */
    globus_object_t    *err;         /* Error type                      */
    globus_io_attr_t attr;           /* I/O attributes of the server    */
    globus_io_handle_t list_handle;  /* Handle for listening socket     */
    globus_io_handle_t conn_handle;  /* Handle for connection socket    */
    globus_size_t nbytes;            /* Number of bytes read or written */

    /* Variables for security functionality */
    globus_io_secure_authorization_data_t auth_data;

    /*  Get port number from the command line */
    if (argc != 2) {
	fprintf(stderr, "ECHOSERV: Wrong number of arguments.\n");
	fprintf(stderr, "Usage: echoserver <port_number>\n");
	fprintf(stderr, "Example: echoserver 1234\n");
	exit(EXIT_FAILURE);
    }
    else {
	port = strtol(argv[1], &endptr, 0);
	if ( *endptr ) {
	    fprintf(stderr, "ECHOSERV: Invalid port number.\n");
	    exit(EXIT_FAILURE);
	}
    }

    /* Activate the Globus I/O module */
    rval = globus_module_activate(GLOBUS_IO_MODULE);
    if (rval != GLOBUS_SUCCESS) {
	printf("ECHOSERV: Globus I/O module activation failed\n");
	exit(EXIT_FAILURE);
    }

    /* Initialize the attributes structure. */
    globus_io_tcpattr_init(&attr);

    if (AUTHENTICATE) {
	/* Use GSS-API authentication */
	rptr = globus_io_attr_set_secure_authentication_mode(&attr, 
             GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI, GSS_C_NO_CREDENTIAL);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOSERV: Setting secure authentication mode failed\n");
	    exit(EXIT_FAILURE);
	}

	/* Initialize the authorization data */
	rptr = globus_io_secure_authorization_data_initialize(&auth_data);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOSERV: Initializing authorization data failed\n");
	    exit(EXIT_FAILURE);
	}
 
	/* Set the authorization mode */
	rptr = globus_io_attr_set_secure_authorization_mode(&attr, 
             GLOBUS_IO_SECURE_AUTHORIZATION_MODE_SELF, &auth_data);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOSERV: Setting authorization mode failed\n");
	    exit(EXIT_FAILURE);
	}

	if (ENCRYPTION) {
	    /* Set the channel mode to use wrapping */
	    rptr = globus_io_attr_set_secure_channel_mode(&attr, 
                 GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
	    if (rptr != GLOBUS_SUCCESS) {
		printf("ECHOSERV: Setting channel mode failed\n");
		exit(EXIT_FAILURE);
	    }

	    /* Use encryption, SAFE instead of PRIVATE is also possible
	       but gives only integrity protection */
	    rptr = globus_io_attr_set_secure_protection_mode(&attr, 
                 GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
	    if (rptr != GLOBUS_SUCCESS) {
		printf("ECHOSERV: Setting encrypted mode failed\n");
		exit(EXIT_FAILURE);
	    }
	}
    }

    /*  Create the listening socket  */
    rptr = globus_io_tcp_create_listener(&port, 2, &attr, &list_handle);
    if (rptr != GLOBUS_SUCCESS) {
	printf("ECHOSERV: Error creating listening socket.\n");
	exit(EXIT_FAILURE);
    }
    
    /*  Enter an infinite loop to respond to client requests and echo input  */
    error = 0;
    while ( 1 ) {
	/*  Wait for a connection */
	rptr = globus_io_tcp_listen(&list_handle);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOSERV: Error calling listen\n");
	    break;
	}

	/*  Connection arrived, accept() it  */
	rptr = globus_io_tcp_accept(&list_handle, &attr, &conn_handle);
	if (rptr != GLOBUS_SUCCESS) {
	    /* Expand the error object and report the type of failure */
	    err = globus_error_get(rptr);
	    if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_AUTHENTICATION_FAILED, 
					 globus_object_get_type(err))) {
		globus_libc_printf("ECHOSERV: Authentication failed\n");
	    }
	    else if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_AUTHORIZATION_FAILED,
					      globus_object_get_type(err))) {
		globus_libc_printf("ECHOSERV: Authorization failed\n");
	    }
	    else if (globus_object_type_match(GLOBUS_IO_ERROR_TYPE_NO_CREDENTIALS,
					      globus_object_get_type(err))) {
		globus_libc_printf("ECHOSERV: Failure: no credentials\n");
	    }
	    else {
		globus_libc_printf("ECHOSERV: Error calling accept\n");
	    }
	    
	    globus_object_free(err);	
	    break;
	}

	/*  Retrieve an input line from the connected socket
	    then simply write it back to the same socket.     */
	rptr = globus_io_read(&conn_handle, buffer, MAX_LINE-1, 1, &nbytes);
	if (rptr != GLOBUS_SUCCESS) {
	    err = globus_error_get(rptr);
	    printf("ECHOSERV: Read failed: %s\n", globus_object_printable_to_string(err));
	    sprintf(buffer, "ECHOSERV: Read failed: %s\n", globus_object_printable_to_string(err));
	    nbytes = strlen(buffer);
	}
	else
	{
	    buffer[nbytes]='\0';
	    printf("Read data: <%s>\n", buffer);
	}

	globus_io_write(&conn_handle, buffer, strlen(buffer), &nbytes);


	/*  Close the connected socket  */
	rptr = globus_io_close(&conn_handle);
	if (rptr != GLOBUS_SUCCESS) {
	    printf("ECHOSERV: Closing the socket failed\n");
	    break;
	}
    }

    if (rptr != GLOBUS_SUCCESS) { /* Error from the loop */
	printf("ECHOSERV: The loop exited due to an error\n");
	error = 1;
    }

    /* Deactivate the Globus I/O module */
    rval = globus_module_deactivate(GLOBUS_IO_MODULE);
    if (rval != GLOBUS_SUCCESS) {
       err = globus_error_get(rptr);
	printf("ECHOSERV: Globus I/O module deactivation failed: %s\n", globus_object_printable_to_string(err));
	exit(EXIT_FAILURE);
    }

    if (error) {
	return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;    
}
